# -*- coding: utf-8 -*-

import sys
import os
from pathlib import Path

# Append parent dir to sys path.
os.sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from cube.io_utils.model_store import ModelMetadata, ModelStore
from cube.io_utils.conll import Dataset
from datetime import datetime
import time
from cube.api import Cube
 
model_tuples = [ #folder, language_code, embedding code
("UD_Afrikaans-AfriBooms","af","af"), 
("UD_Ancient_Greek-PROIEL","grc","grc"), 
("UD_Arabic-PADT","ar","ar"), 
("UD_Armenian-ArmTDP","hy","hy"), 
("UD_Basque-BDT","eu","eu"), 
("UD_Bulgarian-BTB","bg","bg"), 
("UD_Buryat-BDT","bxr","bxr"), 
("UD_Catalan-AnCora","ca","ca"), 
("UD_Chinese-GSD","zh","zh"), 
("UD_Croatian-SET","hr","hr"), 
("UD_Czech-PDT","cs","cs"), 
("UD_Danish-DDT","da","da"), 
("UD_Dutch-Alpino","nl","nl"), 
("UD_English-EWT","en","en"), 
("UD_Estonian-EDT","et","et"), 
("UD_Finnish-TDT","fi","fi"), 
("UD_French-GSD","fr","fr"), 
("UD_Galician-CTG","gl","gl"), 
("UD_German-GSD","de","de"), 
("UD_Gothic-PROIEL","got","got"), 
("UD_Greek-GDT","el","el"), 
("UD_Hebrew-HTB","he","he"), 
("UD_Hindi-HDTB","hi","hi"), 
("UD_Hungarian-Szeged","hu","hu"), 
("UD_Indonesian-GSD","id","id"), 
("UD_Irish-IDT","ga","ga"), 
("UD_Italian-ISDT","it","it"), 
("UD_Japanese-GSD","ja","ja"), 
("UD_Kazakh-KTB","kk","kk"), 
("UD_Korean-GSD","ko","ko"), 
("UD_Kurmanji-MG","kmr","ku"), 
("UD_Latin-ITTB","la","la"), 
("UD_Latvian-LVTB","lv","lv"), 
("UD_North_Sami-Giella","sme","se"), 
("UD_Norwegian-Bokmaal","no_bokmaal","no"), 
("UD_Norwegian-Nynorsk","no_nynorsk","nn"), 
("UD_Old_Church_Slavonic-PROIEL","cu","cu"), 
("UD_Persian-Seraji","fa","fa"), 
("UD_Polish-LFG","pl","pl"), 
("UD_Portuguese-Bosque","pt","pt"), 
("UD_Romanian-RRT","ro","ro"), 
("UD_Russian-SynTagRus","ru","ru"), 
("UD_Serbian-SET","sr","sr"), 
("UD_Slovak-SNK","sk","sk"), 
("UD_Slovenian-SSJ","sl","sl"), 
("UD_Spanish-AnCora","es","es"), 
("UD_Swedish-LinES","sv","sv"), 
("UD_Swedish-Talbanken","sv","sv"), 
("UD_Turkish-IMST","tr","tr"), 
("UD_Ukrainian-IU","uk","uk"), 
("UD_Upper_Sorbian-UFAL","hsb","hsb"), 
("UD_Urdu-UDTB","ur","ur"), 
("UD_Uyghur-UDT","ug","ug"), 
("UD_Vietnamese-VTB","vi","vi")]
    
    
if __name__ == "__main__":
    if len(sys.argv)<2:
        print("Usage: python3 test_all_models.py path-to-UD-repository-root [OPTIONAL-path-to-local-models: where en-1.1 is, etc.]")
        print("Example: python3 test_all_models.py /work/ud-treebanks-v2.2")
        sys.exit(0)
        
    if not os.path.exists("conll18_ud_eval.py"):
        print("Please ensure that conll18_ud_eval.py is located in the same place as test_all_models.py.")
        print("ex: wget http://universaldependencies.org/conll18/conll18_ud_eval.py")
        sys.exit(0)
    from conll18_ud_eval import load_conllu_file, evaluate
    #metrics = ["Tokens", "Sentences", "Words", "UPOS", "XPOS", "UFeats", "AllTags", "Lemmas", "UAS", "LAS", "CLAS", "MLAS", "BLEX"]
    #example usage:     metrics_test = conll_eval(system,gold)
    #                   test_tok_f1, test_ss_f1 = metrics_test["Tokens"].f1*100., metrics_test["Sentences"].f1*100.
    def conll_eval(system_file, gold_file):    
        gold_ud = load_conllu_file(gold_file)
        system_ud = load_conllu_file(system_file)
        return evaluate(gold_ud, system_ud)
    
    if len(sys.argv)==3:
        local_model_path = sys.argv[2]
        print("Using local model path: "+local_model_path)
    else:
        local_model_path = None
        
    model_store_object = ModelStore(local_model_path) 
    local_embeddings_path = model_store_object.embeddings_repository
    local_model_path = model_store_object.disk_path
    
    online_models = model_store_object.list_online_models()
    print("Found {} online models".format(len(online_models)))
    
    local_models = model_store_object.list_local_models()
    print("Found {} local models".format(len(local_models)))
    
    
    model_count = len(online_models)
    
    # step 1. download all models
    for online_model in online_models:        
        model, version = online_model[0], online_model[1]
        if not online_model in local_models:
            print("Downloading {}-{}".format(model,version))
        else:
            print("Model {}-{} is already downloaded.".format(model,version))
            continue
        cube = Cube()
        cube.load(model, version, local_models_repository = local_model_path)
        #cube.load(model)
     
    print("\n\n")
    #for online_model in local_models: #local_models+online_models:
    for online_model in local_models+online_models:
        model, version = online_model[0], online_model[1]
        print("\n\nTesting model {}-{}, @{}".format(model,version, datetime.today()))
        if model == "pl":
            continue
        
        # go run Cube
        print("\t Reading metadata ...")        
        metadata = ModelMetadata()        
        metadata.read(os.path.join(local_model_path,model+"-"+str(version),"metadata.json"))
        
        
        mlanguage = metadata.language
        found = False        
        for elem in model_tuples:            
            if mlanguage in elem[0]:
                ud_language=elem[0]
                ud_language_code=elem[1]
                ud_embedding=elem[2]
                found=True
        if not found:
            print("Could not find model for: metadata.language="+mlanguage)
            sys.exit(0)
        
        # search for already existing processed file
        output_file = ud_language+"."+ud_language_code+"."+model+"."+str(version)+".conllu"
        if os.path.exists(output_file):
           print("\t File already processed, skipping.") 
           continue

        
        test_file_path = os.path.join(sys.argv[1],ud_language)#,ud_language_code+"-ud-test.txt")
        print("\t Searching for test file in "+test_file_path)
        all_files = os.listdir(test_file_path)
        test_file = None
        for f in all_files:
            if ".txt" in f and "-test" in f:
                test_file = f
                break
        
        if not test_file:
            print("Test file not found!")
            sys.exit(0)
        
        
        with open(os.path.join(test_file_path,test_file),"r",encoding="utf8") as f:
            content = f.readlines()
        test_data = ""
        if model == "ja" or model =="zh":
            for line in content:
                test_data+=line.replace("\r","").replace("\n","")
            test_data = test_data.replace(" ","")
        else:
            for line in content:
                test_data+=line.replace("\r","").replace("\n"," ")
            test_data = " ".join(test_data.split())
        
        print("\t Processing {} lines ...".format(len(test_data)))        
        metadata = ModelMetadata()        
        metadata.read(os.path.join(local_model_path,model+"-"+str(version),"metadata.json"))
        local_embeddings_file = os.path.join(local_embeddings_path, metadata.embeddings_file_name)
        cube = Cube(verbose=True)        
        cube.load(model, version, tokenization=True, compound_word_expanding=True, tagging=True,
             lemmatization=True, parsing=True, local_models_repository = local_model_path), 
             #local_embeddings_file = local_embeddings_file) # all enabled, including compound
        start = time.time()        
        sequences = cube(test_data)        
        end = time.time()
        
        ds = Dataset()
        ds.sequences = sequences
        ds.write(output_file)
        
        metrics = conll_eval(output_file, os.path.join(test_file_path,test_file.replace(".txt",".conllu")))
        jsmetrics = {} # apparently the metrics object is not serializable
        for metric in ["Tokens", "Sentences", "Words", "UPOS", "XPOS", "UFeats", "AllTags", "Lemmas", "UAS", "LAS", "CLAS", "MLAS", "BLEX"]:
            jsmetrics[metric] = {}
            jsmetrics[metric]["precision"] = 100 * metrics[metric].precision
            jsmetrics[metric]["recall"] = 100 * metrics[metric].recall
            jsmetrics[metric]["f1"] = 100 * metrics[metric].f1
            jsmetrics[metric]["aligned_accuracy"] = 100 * metrics[metric].aligned_accuracy if metrics[metric].aligned_accuracy is not None else ""
               
        #dump results
        import json
        if not os.path.exists("results.json"):
            results_object = {}
            json.dump(results_object,open("results.json","w"), indent=4, sort_keys=True)
        
        results_object = json.load(open("results.json","r"))
        
       
        elem = {}
        elem["date"] = str(datetime.today())
        elem["metrics"] = jsmetrics
        elem["test_file"] = os.path.join(test_file_path,test_file)
        elem["test_time"] = end - start
        results_object[model+"-"+str(version)] = elem
        json.dump(results_object,open("results.json","w"), indent=4, sort_keys=True)
        
    print("\n\nFinished everything.")


import sys

sys.path.append('')
from cube.io_utils.misc import fopen


class Document:
    """
    Document (document)

    A document (Document) is a collection of sentences. The structures inside are compatible with CONLL-U format. See https://universaldependencies.org/format.html for more details

    Example usage:
        doc = Document(filename='corpus/ud-treebanks-v2.7/UD_Romanian-RRT/ro_rrt-ud-dev.conllu')
    """

    def __init__(self, filename: str = None, lang_id: int = 0):
        """
        Create a new Document instance.

        You can pass optional arguments filename and lang_id if you want to load a UD-Style document
        """
        self.sentences = []
        if filename is not None:
            self.load(filename, lang_id)

    def load(self, filename: str, lang_id: int = 0):
        """
        Load a CONLL-U file into the document

        Params:
            filename - mandatory parameter
            lang_id - optional parameter (default value=0)

        Example usage:
            doc = Document()
            doc.load('corpus/ud-treebanks-v2.7/UD_Romanian-RRT/ro_rrt-ud-dev.conllu', lang_id=1000)

        """
        in_sequence = False
        f = fopen(filename, 'r')
        seq = []
        cnt = 0
        for line in f.readlines():
            line = line.replace("\n", "")
            line = line.replace("\r", "")
            cnt += 1
            # if cnt == 100:
            #     break
            if (not line.startswith("#") or in_sequence) and line != '':
                parts = line.split("\t")

                s = Word(parts[0], parts[1], parts[2], parts[3], parts[4], parts[5], parts[6], parts[7],
                         parts[8], parts[9])
                seq.append(s)
                in_sequence = True
            elif line == "":
                in_sequence = False
                if len(seq) > 0:
                    self.sentences.append(Sentence(sequence=seq, lang_id=lang_id))
                seq = []
        f.close()

    def text(self):
        self.__repr__()

    def __repr__(self):
        return '\n\n'.join([str(s) for s in self.sentences])


class Sentence:
    """
    A sentence is a collection of tokens and words. This class is not meant to be initialized outside of the Document structure
    """

    def __init__(self, sequence=None, lang_id=0, text=None):
        self.doc = None
        self.tokens = []
        self.words = []

        self.lang_id = lang_id
        skip = 0
        t = None
        if sequence is not None:
            for w in sequence:
                if '.' in str(w.index):
                    continue
                if skip == 0:
                    t = Token(index=w.index, text=w.word, space_after=(not 'spaceafter=no' in w.space_after.lower()),
                              words=[])
                    self.tokens.append(t)
                    if w.is_compound_entry:
                        parts = w.index.split('-')
                        skip = int(parts[1]) - int(parts[0]) + 1
                    else:
                        skip = 1

                if not w.is_compound_entry:
                    skip -= 1
                    w.token = t
                    t.words.append(w)
                    self.words.append(w)

        if text is None:
            self.text = self._detokenize()
        else:
            self.text = text

    def _detokenize(self):
        s = []
        for t in self.tokens:
            s.append(t.text)
            if t.space_after:
                s.append(' ')
        return ''.join(s)

    def __repr__(self):
        return '\n'.join([str(t) for t in self.tokens])


class Token:
    """
    A token contains a list of composing words. Except for Multiword Tokens, the list of words will contain a single element.
    This class is not meant to be initialized outside the Document/Sentence structures.
    """

    def __init__(self, index=0, text: str = '', words=[], space_after=True):
        self.index = index
        self.text = text
        self.words = words
        self.other = None  # ner, sentiment, etc
        self.space_after = space_after

    def __repr__(self):
        if not self.space_after:
            spa = 'SpaceAfter=No'
        else:
            spa = '_'
        head = ''
        if len(self.words) > 1:
            head = "\t".join([str(self.index), self.text,
                              '_', '_',
                              '_', '_', '_', '_', '_', spa]) + '\n'
        return head + '\n'.join([str(w) for w in self.words])


def _int_try_parse(value):
    try:
        return int(value), False
    except ValueError:
        return value, True


class Word:
    """
    Structure to hold CONLL-U style metadata for words. See https://universaldependencies.org/format.html for more details about the format
    """

    def __init__(self, index, word: str, lemma: str, upos: str, xpos: str, attrs: str, head, label: str, deps: str,
                 space_after: str, token: Token = None):
        self.index, self.is_compound_entry = _int_try_parse(index)
        self.word = word
        self.lemma = lemma
        self.upos = upos
        self.xpos = xpos
        self.attrs = attrs
        self.head, _ = _int_try_parse(head)
        self.label = label
        self.deps = deps
        self.space_after = space_after
        self.parent = token
        self.emb = None

    def __repr__(self):
        return "\t".join([str(self.index), self.word if isinstance(self.word, str) else self.word.encode('utf-8'),
                          self.lemma if isinstance(self.lemma, str) else self.lemma.encode('utf-8'), self.upos,
                          self.xpos, self.attrs, str(self.head), self.label, self.deps, self.space_after])


if __name__ == '__main__':
    print("test")
    doc = Document(filename='corpus/ud-treebanks-v2.7/UD_French-GSD/fr_gsd-ud-dev.conllu')
    print(doc)



import json
import sys
import optparse

sys.path.append('')


def _is_complete_corpus(filename):
    from cube.io_utils.conll import Dataset
    dataset = Dataset()
    dataset.load_language(filename, 0)
    ok = False
    for entry in dataset.sequences[0][0]:
        word = entry.word.replace('_', '').strip()
        if word != '':
            ok = True
            break
    return ok


def _get_list_of_folders_containing(dirName, pattern):
    import os
    listOfFile = os.listdir(dirName)
    allFiles = list()
    # Iterate over all the entries
    for entry in listOfFile:
        # Create full path
        fullPath = os.path.join(dirName, entry)
        if os.path.isdir(fullPath) and pattern in fullPath:
            allFiles.append(entry)
    return allFiles


def _get_list_of_files(dirName):
    import os
    listOfFile = os.listdir(dirName)
    allFiles = list()
    # Iterate over all the entries
    for entry in listOfFile:
        # Create full path
        fullPath = os.path.join(dirName, entry)
        if os.path.isfile(fullPath):
            allFiles.append(fullPath)
    return allFiles


def _get_file(train_base, language, type):
    import os.path as path
    pth = path.join(train_base, language)
    files = _get_list_of_files(pth)
    for file in files:
        if file.endswith('-{0}.conllu'.format(type)):
            return file
    return ''


def _get_lang_id(file):
    parts = file.split('/')
    lang_id = parts[-1].split('-')[0]
    return lang_id


def _process_single(params):
    lines = open(params.input_file).readlines()
    combined = []
    for line in lines:
        line = line.strip()
        train_file = _get_file(params.train_base, line, 'train')
        dev_file = _get_file(params.train_base, line, 'dev')
        if dev_file == '':
            dev_file = _get_file(params.train_base, line, 'test')
        if train_file != '' and dev_file != '':
            lang_id = _get_lang_id(train_file)
            if _is_complete_corpus(dev_file):
                combined.append([lang_id, train_file, dev_file])
            else:
                sys.stdout.write('Removing incomplete languge: "{0}"\n'.format(lang_id))

    json.dump(combined, open(params.output_file, 'w'))


def _process_multi(params):
    families = {'Germanic': 10,
                'Afro-Asiatic': 7,
                'Albanian': 1,
                'Greek': 2,
                'Armenian': 1,
                'Mande': 1,
                'Basque': 1,
                'Slavic': 13,
                'Indic': 6,
                'Celtic': 4,
                'Mongolic': 1,
                'Sino-Tibetan': 3,
                'Romance': 8,
                'Uralic': 11,
                'Austronesian': 1,
                'Japanese': 1,
                'Turkic': 3,
                'Korean': 1,
                'Iranian': 2,
                'Latin': 1,
                'Baltic': 2,
                'Tupian': 1,
                'Creole': 1,
                'Dravidian': 1,
                'Tai-Kadai': 1,
                'Austro-Asiatic': 1,
                'Pama-Nyungan': 1,
                'Niger-Congo': 1}

    family2dataset = {k: [] for k in families}
    lines = open(params.input_file).readlines()
    added_folders = {}
    for line in lines:
        line = line.strip()
        parts = line.split('  ')
        ltypes = parts[1].replace(', ', ',').split(',')
        family = ''
        for ltype in ltypes:
            if ltype in family2dataset:
                family = ltype
                break
        if family == '':
            print("Unable to process " + line)
        else:
            language_name = parts[0].split(' ')[0]
            all_folders = _get_list_of_folders_containing(params.train_base, language_name)
            if len(all_folders) == 0:
                print("Unable to find training data for " + language_name)
            else:
                for folder in all_folders:
                    if folder in added_folders:
                        continue

                    added_folders[folder] = 1
                    train_file = _get_file(params.train_base, folder, 'train')
                    dev_file = _get_file(params.train_base, folder, 'dev')
                    if dev_file == '':
                        dev_file = _get_file(params.train_base, folder, 'test')
                    if train_file != '' and dev_file != '':
                        lang_id = _get_lang_id(train_file)
                        if _is_complete_corpus(dev_file):
                            family2dataset[family].append([lang_id, train_file, dev_file])
                        else:
                            sys.stdout.write('Removing incomplete languge: "{0}"\n'.format(lang_id))
    for fam in family2dataset:
        out_file = '{1}/{0}.json'.format(fam.lower().replace(' ', '_'), params.output_file)
        json.dump(family2dataset[fam], open(out_file, 'w'), indent=4)

    # list of all languages
    all = []
    for fam in family2dataset:
        for entry in family2dataset[fam]:
            all.append(entry)

    out_file = '{0}/all.json'.format(params.output_file)
    json.dump(all, open(out_file, 'w'), indent=4)


def _process_langs(params):
    def _extract_lang_and_flavour(name):
        name = name[3:]
        parts = name.split('-')
        lang = parts[0]
        flavour = parts[1]
        return lang, flavour

    lines = open(params.input_file).readlines()
    lang2flavour = {}
    for line in lines:
        dataset_name = line.strip().split(' ')[-1]

        train_file = _get_file(params.train_base, dataset_name, 'train')
        dev_file = _get_file(params.train_base, dataset_name, 'dev')
        if dev_file == '':
            dev_file = _get_file(params.train_base, dataset_name, 'test')
        if train_file != '' and dev_file != '':
            lang_id = _get_lang_id(train_file)
            if _is_complete_corpus(dev_file):
                lang, flavour = _extract_lang_and_flavour(dataset_name)
                if lang not in lang2flavour:
                    lang2flavour[lang] = [(flavour, lang_id, train_file, dev_file)]
                else:
                    lang2flavour[lang].append((flavour, lang_id, train_file, dev_file))
    for lang in lang2flavour:
        trainset = []
        for item in lang2flavour[lang]:
            # flavour = item[0]
            lang_id = item[1]
            train_file = item[2]
            dev_file = item[3]
            trainset.append([lang_id, train_file, dev_file])
        output_file = '{0}/{1}.json'.format(params.output_file, lang.lower())
        f = open(output_file, 'w')
        json.dump(trainset, f, indent=4)
        f.close()
        sys.stdout.write('python3 cube/networks/model.py --batch-size=32 --device=cuda:0 --store=data/lemmatizer-{0} --train=examples/multilanguage/languages/{0}.json\n'.format(lang.lower()))


if __name__ == '__main__':
    parser = optparse.OptionParser()
    parser.add_option('--input-file', action='store', dest='input_file')
    parser.add_option('--output-file', action='store', dest='output_file')
    parser.add_option('--train-base', action='store', dest='train_base')
    parser.add_option('--type', action='store', dest='type', choices=['single', 'family', 'language'])

    (params, _) = parser.parse_args(sys.argv)

    if params.input_file and params.output_file and params.train_base:
        if params.type == 'single':
            _process_single(params)
        elif params.type == 'family':
            _process_multi(params)
        elif params.type == 'language':
            _process_langs(params)
    else:
        parser.print_help()
